/*
 * Copyright (c) 2012 - Network Appliance Inc.
 *
 */

/*********************************************************************
 * MV88E6176 driver module
 *
 *********************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/mutex.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/cacheflush.h>
#include <asm/ioctl.h>

#include <linux/kobject.h>
#include <linux/string.h>
#include <linux/sysfs.h>
#include <linux/module.h>

#include "sw_sysfs.h"
#include "mv88e6176_regdefs.h"
#include "mv88e6176.h"
#include "mdcio.h"
#include "enet_switch.h"


/*****Prototye Functions*************************/
static int    mv88e6176_init (void );
static void   mv88e6176_exit (void);
static int    mv88e6176_open(struct inode * inode, struct file * filp) ;
static long   mv88e6176_ioctl(struct file *file, unsigned int cmd, unsigned long arg) ;


// An operation is defineds as a set of commands that need to be executed aomically.
// If there is a function that has a bunch of commands wrap it with mv88e6176_operation_lock.
// These functions can be called from sysfs or from ioctls.
static DEFINE_MUTEX(mv88e6176_operation_lock) ;

// Register access lock is applied to read/write for port and phy registers only.
// Don't use this lock anywhere else in code, except for the functions that has it.
// Make sure taken lock is not taken again.
static DEFINE_MUTEX(mv88e6176_register_access_lock) ;

static unsigned int g_register_access_lock_cnt = 0 ;
static unsigned int g_register_access_unlock_cnt = 0 ;

static unsigned int g_operation_lock_cnt = 0 ;
static unsigned int g_operation_unlock_cnt = 0 ;

static unsigned int g_debug_print = 0 ;

static unsigned int port_to_phy_dev_addr_map[MRV_88E6176_MAX_PORTS] = {
    MRV_88E6176_PHY0_DEV_ADDR,      // port 0 is associated to phy 0
    MRV_88E6176_PHY1_DEV_ADDR,      // port 1 is associated to phy 1
    MRV_88E6176_PHY2_DEV_ADDR,      // port 2 is associated to phy 2
    MRV_88E6176_PHY3_DEV_ADDR,      // port 3 is associated to phy 3
    MRV_88E6176_PHY4_DEV_ADDR,      // port 4 is associated to phy 4
    MRV_88E6176_PHY_UNKNOWN_ADDR ,  // port 5 has no phy.
    MRV_88E6176_PHY_UNKNOWN_ADDR    // port 6 has no phy.
} ;

static int mv88e6176_read_reg(unsigned int phyaddr, unsigned int regnum, unsigned int *data)
{
    int retval = 0 ;
    if ( data == NULL )
        return -EINVAL ;

    retval = mdcio_read_reg(0, phyaddr, regnum, data);
    mv88e6176_log(g_debug_print, "%s::%d(B) retval = %d DA = 0x%x, RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, phyaddr,regnum,*data ) ;
    return retval ;
}

static int mv88e6176_write_reg(unsigned int phyaddr, unsigned int regnum, unsigned int data)
{
    int retval = 0 ;
    retval = mdcio_write_reg(0, phyaddr, regnum, data) ;
    mv88e6176_log(g_debug_print, "%s::%d(B) retval = %d DA = 0x%x, RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, phyaddr,regnum,data ) ;
    return retval ;
}

static int mv88e6176_validate_port_glob_params(unsigned int dev_addr, unsigned int reg_addr)
{
    if ( ( ( dev_addr  >=  MRV_88E6176_PORT_START_DEV_ADDR  )  && ( dev_addr  <=  MRV_88E6176_GLOBAL2_DEV_ADDR) ) &&
         ( ( reg_addr  >=  MRV_88E6176_PORT_REG_START_ADDR  )  && ( reg_addr  <=  MRV_88E6176_PORT_REG_END_ADDR ) ) )
        return 0 ;
    else {
        return -EINVAL ;
    }
}

static int mv88e6176_validate_phy_params(unsigned int dev_addr, unsigned int page_addr, unsigned int reg_addr)
{
    if ( ( dev_addr  >=  MRV_88E6176_PHY_START_DEV_ADDR  )  && ( dev_addr  <=  MRV_88E6176_PHY_END_DEV_ADDR) &&
         ( page_addr >=  MRV_88E6176_PHY_PAGE_START_ADDR )  && ( page_addr <=  MRV_88E6176_PHY_PAGE_END_ADDR ) &&
         ( reg_addr  >=  MRV_88E6176_PHY_REG_START_ADDR  )  && ( reg_addr  <  MRV_88E6176_PHY_REG_END_ADDR ) ) {
        return 0 ;
    }
    else {
        return -EINVAL ;
    }
}

static int mv88e6176_indirect_send_op(int dev_addr, int reg_addr, unsigned int data, unsigned short op)
{
    int retval = 0 ;
    int rval = 0 ;
    unsigned char iter = 0 ;
    unsigned char busybit = 1 ;

    // Indirect writes are done only for phy registers. So validate for phy addresses.
    // Validation is for dev_addr and reg_addr, so let the page_addr be a correct value of 0.
    retval = mv88e6176_validate_phy_params(dev_addr, 0 , reg_addr) ;
    if ( retval == 0 ) {

        // Write SMI_DATA register with value.
        retval = mv88e6176_write_reg(MRV_88E6176_GLOBAL2_DEV_ADDR, MRV_88E6176_GLOBAL2_SMI_PHY_DATA_REG_ADDR, data) ;
        if ( retval == 0 ) {

            // Write to SMI_CONTROL global register giving the operation to perform.
            retval = mv88e6176_write_reg(MRV_88E6176_GLOBAL2_DEV_ADDR, MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_ADDR, op) ;
            if ( retval == 0 ) {
                // Wait until the SMIBusy bit becomes 0.
                while ( ( iter < SMI_BUSY_BIT_ITER_MAX ) && ( busybit == 1 ) && ( retval == 0 ) ) {
                    retval = mv88e6176_read_reg(MRV_88E6176_GLOBAL2_DEV_ADDR, MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_ADDR,&rval) ;
                    busybit = ( rval & MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIBUSY_MASK ) >> MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIBUSY_SHIFT ;
                    iter++ ;
                    udelay(SMI_BUSY_BIT_DELAY) ;
                    rval = 0 ;
                }

                if ( ( busybit ) || ( retval != 0 ) ) {
                    printk("%s:: busybit is still set, or retval != 0, rval = %d, retval = 0x%x \n", __FUNCTION__, rval, retval ) ;
                    return -EINVAL  ; // Return error here.
                }
            }
        }
    }
    return retval ;
}

static int mv88e6176_indirect_read_reg(int dev_addr, int reg_addr, int *data)
{
    int retval = 0 ;
    unsigned short op = 0 ;
    int rval = 0 ;

    if ( data == NULL )
        return -EINVAL ;

    op = ( 1 << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIBUSY_SHIFT ) |
         ( MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIMODE_VAL_CLAUSE_22 <<MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIMODE_SHIFT ) |
         ( MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIOP_VAL_SMIMODE_1_READ_DATA_REG << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIOP_SHIFT ) |
         ( dev_addr << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_DEV_ADDR_SHIFT ) |
         ( reg_addr << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_REG_ADDR_SHIFT ) ;

    // data arg is 0, because we want to clear up content on data register.
    retval = mv88e6176_indirect_send_op(dev_addr, reg_addr, 0, op) ;
    if ( retval == 0 ) {
        // Read from SMI_DATA global register.
        retval = mv88e6176_read_reg( MRV_88E6176_GLOBAL2_DEV_ADDR, MRV_88E6176_GLOBAL2_SMI_PHY_DATA_REG_ADDR,&rval) ;
        if ( retval == 0 ) {
            *data = rval ;
        }
    }
    return retval ;
}

static int mv88e6176_indirect_write_reg(int dev_addr, int reg_addr, int data)
{
    unsigned short op = 0 ;

    // Write to SMI_CONTROL global register giving the instruction to write the dev addr's register.
    op = ( 1 << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIBUSY_SHIFT ) |
         ( MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIMODE_VAL_CLAUSE_22 <<MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIMODE_SHIFT ) |
         ( MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIOP_VAL_SMIMODE_1_WRITE_DATA_REG << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_SMIOP_SHIFT ) |
         ( dev_addr << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_DEV_ADDR_SHIFT ) |
         ( reg_addr << MRV_88E6176_GLOBAL2_SMI_PHY_COMMAND_REG_REG_ADDR_SHIFT ) ;

    return ( mv88e6176_indirect_send_op(dev_addr, reg_addr, data, op) ) ;
}

void mv88e6176_get_debug_print(unsigned int *val) 
{
    *val = g_debug_print ;
}

void mv88e6176_set_debug_print(unsigned int val) 
{
    g_debug_print = val ;
}

int mv88e6176_get_lock_counters(unsigned int *reg_access_lock_cntr, unsigned int *reg_access_unlock_cntr, unsigned int *operation_lock_cntr, unsigned int *operation_unlock_cntr)
{
    if ( ( reg_access_lock_cntr == NULL ) || ( reg_access_unlock_cntr == NULL ) || ( operation_lock_cntr == NULL ) || ( operation_unlock_cntr == NULL ) ) {
        return -EINVAL ;
    }

    *reg_access_lock_cntr = g_register_access_lock_cnt ;
    *reg_access_unlock_cntr = g_register_access_unlock_cnt ;
    *operation_lock_cntr  = g_operation_lock_cnt ;
    *operation_unlock_cntr  = g_operation_unlock_cnt ;

    return 0 ;
}

int mv88e6176_set_lock_counters(unsigned int reg_access_lock_cntr, unsigned int reg_access_unlock_cntr, unsigned int operation_lock_cntr, unsigned int operation_unlock_cntr)
{
    g_register_access_lock_cnt = reg_access_lock_cntr ;
    g_register_access_unlock_cnt = reg_access_unlock_cntr ;
    g_operation_lock_cnt = operation_lock_cntr ;
    g_operation_unlock_cnt = operation_unlock_cntr ;
    return 0 ;
}


int mv88e6176_port_glob_read_reg(unsigned int dev_addr, unsigned int reg_addr , unsigned int *value)
{
    int retval = 0 ;

    if ( value == NULL ) 
        return -EINVAL ;

    mv88e6176_lock(mv88e6176_register_access_lock, flags, g_register_access_lock_cnt, "mv88e6176_port_glob_read_reg", &g_register_access_lock_name[0])  ;

     *value = 0 ;
     mv88e6176_log(g_debug_print, "%s::%d(B) retval = %d PG_Read_Reg DA = 0x%x, RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, dev_addr,reg_addr,*value ) ;
    // Validate register address ranges and return error if needed.
    retval = mv88e6176_validate_port_glob_params(dev_addr, reg_addr) ;
    if ( retval == 0 ) {
        retval = mv88e6176_read_reg(dev_addr, reg_addr,value) ;
        if ( retval != 0 ) {
            printk("(F)%s read_reg retval = 0x%x \n",__FUNCTION__, retval ) ;
        }
        mv88e6176_log(g_debug_print, "%s::%d(E) retval = %d PG_Read_Reg DA = 0x%x, RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, dev_addr,reg_addr,*value ) ;
        mv88e6176_log(g_debug_print, "************\n") ;
    } else {
        printk("(F)%s validate port params dev_addr = 0x%x , reg_addr = 0x%x \n",__FUNCTION__, dev_addr, reg_addr) ;
    }

    mv88e6176_unlock(mv88e6176_register_access_lock, flags, g_register_access_unlock_cnt, "mv88e6176_port_glob_read_reg", &g_register_access_lock_name[0])  ;

    return retval ;
}

int mv88e6176_port_glob_write_reg(unsigned int dev_addr, unsigned int reg_addr , unsigned int value)
{
    int retval = 0 ;

    mv88e6176_lock(mv88e6176_register_access_lock, flags, g_register_access_lock_cnt, "mv88e6171_port_glob_write_reg", &g_register_access_lock_name[0])  ;

     mv88e6176_log(g_debug_print, "%s::%d(B) retval = %d PG_Write_Reg DA = 0x%x, RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval , dev_addr,reg_addr,value ) ;
    // After obtaining arguments, read the register and keep it ready so that a cat on attribute will show the value.
    retval = mv88e6176_validate_port_glob_params(dev_addr, reg_addr) ;
    if ( retval == 0 ) {
        retval = mv88e6176_write_reg(dev_addr, reg_addr,value) ;
        if ( retval != 0 ) {
            printk("(F)%s write_reg  \n",__FUNCTION__ ) ;
        }
        mv88e6176_log(g_debug_print, "%s::%d(E) retval = %d PG_Write_Reg DA = 0x%x, RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, dev_addr,reg_addr,value ) ;
        mv88e6176_log(g_debug_print, "************\n") ;
    } else {
        printk("(F)%s validate port params \n",__FUNCTION__ ) ;
    }

    mv88e6176_unlock(mv88e6176_register_access_lock, flags, g_register_access_unlock_cnt, "mv88e6176_port_glob_write_reg", &g_register_access_lock_name[0])  ;

    return retval ;
}

static int mv88e6176_set_pagenum(int dev_addr, int page_num)
{
    int retval = 0 ;
    unsigned int wval = 0 ;
    unsigned int data = 0 ;

    // Pass a valid value for register address.
    retval = mv88e6176_validate_phy_params(dev_addr, page_num, 0 ) ;
    if ( retval == 0 ) {
        // Read page register.
        retval = mv88e6176_indirect_read_reg(dev_addr, MRV_88E6176_PHY_DEV_PAGE_ADDR_REG , &data) ;
        if ( retval == 0 ) {
            // Set page number.
            wval = (data & ~MRV_88E6176_PHY_DEV_PAGE_ADDR_REG_PAGE_SELECT_MASK ) | page_num ;

            // Write the page number back.
            retval = mv88e6176_indirect_write_reg(dev_addr, MRV_88E6176_PHY_DEV_PAGE_ADDR_REG, wval) ;
        } else  {
            retval = -EINVAL ;
        }
    } else {
        retval = -EINVAL ;
    }

    return retval ;
}

int mv88e6176_phy_read_reg(unsigned int dev_addr, unsigned int page_addr, unsigned int reg_addr, unsigned int *value )
{
    int retval = 0 ;

    if ( value == NULL ) 
        return -EINVAL ;

    mv88e6176_lock(mv88e6176_register_access_lock, flags, g_register_access_lock_cnt, "mv88e6176_phy_read_reg", &g_register_access_lock_name[0])  ;

    *value = 0 ;
    mv88e6176_log(g_debug_print, "%s::%d(B) retval = %d PHY_Read_Reg DA = 0x%x, PA = 0x%x RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, dev_addr, page_addr, reg_addr,*value ) ;
    // Validate register address ranges and return error if needed.
    retval = mv88e6176_validate_phy_params(dev_addr, page_addr, reg_addr) ;
    if ( retval == 0 ) {
        // Set page number, perform an indirect read.
        retval = mv88e6176_set_pagenum(dev_addr, page_addr) ;
        if ( retval == 0 ) {
            // Now read the actual register, perform an indirect read.
            retval = mv88e6176_indirect_read_reg(dev_addr, reg_addr,value) ;
            if ( retval != 0 ) {
                printk("(F)%s indirect read reg",__FUNCTION__ ) ;
            }
            mv88e6176_log(g_debug_print, "%s::%d(E) retval = %d PHY_Read_Reg DA = 0x%x, PA = 0x%x RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, dev_addr, page_addr, reg_addr,*value ) ;
            mv88e6176_log(g_debug_print, "++++++++++++ \n") ;
        } else {
            printk("(F)%s set pagenum ",__FUNCTION__ ) ;
        }
    } else {
       printk("(F)%s validate params dev_addr = 0x%x, page_num = 0x%x, reg_addr = 0x%x \n",__FUNCTION__,dev_addr, page_addr, reg_addr  ) ;
    }

    // mutex_unlock(&mv88e6176_register_access_lock) ;
    mv88e6176_unlock(mv88e6176_register_access_lock, flags, g_register_access_unlock_cnt, "mv88e6176_phy_read_reg", &g_register_access_lock_name[0])  ;

    return retval ;
}

int mv88e6176_phy_write_reg(unsigned int dev_addr, unsigned int page_addr, unsigned int reg_addr, unsigned int value )
{
    int retval = 0 ;

    mv88e6176_lock(mv88e6176_register_access_lock, flags, g_register_access_lock_cnt, "mv88e6176_phy_write_reg", &g_register_access_lock_name[0])  ;

    mv88e6176_log(g_debug_print, "%s::%d(B) retval = %d PHY_Write_Reg DA = 0x%x, PA = 0x%x RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, dev_addr, page_addr, reg_addr,value ) ;
    // Validate register address ranges and return error if needed.
    retval = mv88e6176_validate_phy_params(dev_addr, page_addr, reg_addr) ;
    if ( retval == 0 ) {
        // Set page number, perform an indirect read.
        retval = mv88e6176_set_pagenum(dev_addr, page_addr) ;
        if ( retval == 0 ) {
            // Now write to actual register, perform an indirect write.
            retval = mv88e6176_indirect_write_reg(dev_addr, reg_addr,value) ;
            if ( retval != 0 ) {
                printk("(F)%s indirect write reg",__FUNCTION__ ) ;
            }
            mv88e6176_log(g_debug_print, "%s::%d(E) retval = %d PHY_Write_Reg DA = 0x%x, PA = 0x%x RA = 0x%x, val = 0x%x \n", __FUNCTION__, g_register_access_lock_cnt, retval, dev_addr, page_addr, reg_addr,value ) ;
            mv88e6176_log(g_debug_print, "++++++++++++ \n") ;
        } else {
            printk("(F)%s set pagenum ",__FUNCTION__ ) ;
        }
    } else {
       printk("(F)%s validate params dev_addr = 0x%x, page_num = 0x%x, reg_addr = 0x%x \n",__FUNCTION__,dev_addr, page_addr, reg_addr  ) ;
    }

    // mutex_unlock(&mv88e6176_register_access_lock) ;
    mv88e6176_unlock(mv88e6176_register_access_lock, flags, g_register_access_unlock_cnt, "mv88e6176_phy_write_reg", &g_register_access_lock_name[0])  ;

    return retval ;
}

unsigned int speed_val[] = { SPEED_10_MBPS ,    // 10 Mbps
                             SPEED_100_MBPS ,   // 100 Mbps
                             SPEED_1000_MBPS ,  // 1000 Mbps
                             SPEED_UNKNOWN_MBPS // RSVD. Means in the register speed is not set.
                           } ;

int mv88e6176_read_port_status(int port_num, PORT_DETS_T pd , unsigned int *data)
{
    int retval = 0 ;
    unsigned int rval = 0 ;
    unsigned int temp = 0 ;

    // mutex_lock(&mv88e6176_operation_lock) ;
    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_read_port_status" , &g_operation_lock_name[0])  ;

    if ( ( port_num >= 0 ) && ( port_num <  MRV_88E6176_MAX_PORTS ) && ( data != NULL ) ) {
        if ( pd == FLOWCTRL ) { 
            // Read from a phy register.
            // Read reg 4.
            retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                          MRV_88E6176_PHY_COPPER_AUTO_NEGOTIATION_ADVERTISEMENT_REG , &rval ) ;
            // printk("REGISTER 4 value = 0x%x \n", rval ) ;		
            if ( retval == 0 ) { 
                *data = (rval & MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_PAUSE_MASK) >> MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_PAUSE_SHIFT ;
            } else { 
                retval = -EINVAL ;
            }
        } else { 
            retval = mv88e6176_port_glob_read_reg((port_num + MRV_88E6176_PORT_START_DEV_ADDR), MRV_88E6176_PORT_STATUS_REG_ADDR, &rval)  ;
            if ( retval == 0 ) {
                switch (pd) {
                    case SPEED :
                    {
                        temp = (rval & MRV_88E6176_PORT_STATUS_REG_SPEED_MASK) >> MRV_88E6176_PORT_STATUS_REG_SPEED_SHIFT ;
                        *data = speed_val[temp] ;
                        break ;
                    }
                    case DUPLEX :
                    {
                        *data = (rval & MRV_88E6176_PORT_STATUS_REG_DUPLEX_MASK) >> MRV_88E6176_PORT_STATUS_REG_DUPLEX_SHIFT ;
                        break ;
                    }
                    case LINK_STATUS :
                    {
                        *data = (rval & MRV_88E6176_PORT_STATUS_REG_LINK_STATUS_MASK) >> MRV_88E6176_PORT_STATUS_REG_LINK_STATUS_SHIFT ;
                        break ;
                    }
                    default :
                    {
                        printk("%s Invalid type %d \n", __FUNCTION__, pd ) ;
                        break ;
                    }
                } // End of switch
            } else {
                retval = -EINVAL ;
            }
        }
    } else  {
        retval = -EINVAL ;
    }

    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_read_port_status" , &g_operation_lock_name[0])  ;
    return retval ;
}


// There are 3 speeds ( 10, 100, 1000 ) and 2 duplexs ( half, full ) 
// so array is 3x2 that is 6 elements .

struct speed_duplex_op { 
    unsigned int speed ; 
    duplex_t duplex ; 
    speed_duplex_op_ctrl_t ctrl ;            
} ; 

typedef enum {
    SPEED_10_HALF_DUPLEX = 0 , 
    SPEED_10_FULL_DUPLEX     , 
    SPEED_100_HALF_DUPLEX    , 
    SPEED_100_FULL_DUPLEX    ,
    SPEED_1000_HALF_DUPLEX   , 
    SPEED_1000_FULL_DUPLEX   , 
} speed_duplex_arr_slot_t ;

#define MAX_SPEED_DUPLEX_OPS 6

// Array initialized with speed, duplex, and no op.
static struct speed_duplex_op speed_duplex_op_arr[MAX_SPEED_DUPLEX_OPS] = { 
    { SPEED_10_MBPS   , HALF_DUPLEX, SD_OP_NO_OP } , // 10 MBPS, Half duplex, no op.
    { SPEED_10_MBPS   , FULL_DUPLEX, SD_OP_NO_OP } , // 10 MBPS, Half duplex, no op.
    { SPEED_100_MBPS  , HALF_DUPLEX, SD_OP_NO_OP } , // 10 MBPS, Half duplex, no op.
    { SPEED_100_MBPS  , FULL_DUPLEX, SD_OP_NO_OP } , // 10 MBPS, Half duplex, no op.
    { SPEED_1000_MBPS , HALF_DUPLEX, SD_OP_NO_OP } , // 10 MBPS, Half duplex, no op.
    { SPEED_1000_MBPS , FULL_DUPLEX, SD_OP_NO_OP } , // 10 MBPS, Half duplex, no op.
};

// SPEED_10_HALF_DUPLEX_ARR
static struct speed_duplex_op speed_10_half_duplex_op_arr[MAX_SPEED_DUPLEX_OPS] = { 
    { SPEED_10_MBPS   , HALF_DUPLEX, SD_OP_ADVERTISE } , // 10 MBPS, Half duplex, advertise op.
    { SPEED_10_MBPS   , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise.
    { SPEED_100_MBPS  , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise.
    { SPEED_100_MBPS  , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise.
};
 
// SPEED_10_FULL_DUPLEX_ARR
static struct speed_duplex_op speed_10_full_duplex_op_arr[MAX_SPEED_DUPLEX_OPS] = { 
    { SPEED_10_MBPS   , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, advertise op.
    { SPEED_10_MBPS   , FULL_DUPLEX, SD_OP_ADVERTISE } , // 10 MBPS, Full duplex, not advertise.
    { SPEED_100_MBPS  , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 100 MBPS, Half duplex, not advertise.
    { SPEED_100_MBPS  , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 100 MBPS, Full duplex, not advertise.
    { SPEED_1000_MBPS , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, Full duplex, not advertise.
};
 
// SPEED_100_HALF_DUPLEX_ARR
static struct speed_duplex_op speed_100_half_duplex_op_arr[MAX_SPEED_DUPLEX_OPS] = { 
    { SPEED_10_MBPS   , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise .
    { SPEED_10_MBPS   , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise.
    { SPEED_100_MBPS  , HALF_DUPLEX, SD_OP_ADVERTISE } ,     // 100 MBPS, Half duplex, advertise.
    { SPEED_100_MBPS  , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 100 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, Half duplex, not advertise.
};
 
// SPEED_100_FULL_DUPLEX_ARR
static struct speed_duplex_op speed_100_full_duplex_op_arr[MAX_SPEED_DUPLEX_OPS] = { 
    { SPEED_10_MBPS   , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise .
    { SPEED_10_MBPS   , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, full duplex, not advertise.
    { SPEED_100_MBPS  , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 100 MBPS, Half duplex, advertise.
    { SPEED_100_MBPS  , FULL_DUPLEX, SD_OP_ADVERTISE } ,     // 100 MBPS, full duplex, not advertise.
    { SPEED_1000_MBPS , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, full duplex, not advertise.
};
 
// SPEED_1000_HALF_DUPLEX_ARR
static struct speed_duplex_op speed_1000_half_duplex_op_arr[MAX_SPEED_DUPLEX_OPS] = { 
    { SPEED_10_MBPS   , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise .
    { SPEED_10_MBPS   , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise.
    { SPEED_100_MBPS  , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } ,     // 100 MBPS, Half duplex, advertise.
    { SPEED_100_MBPS  , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 100 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , HALF_DUPLEX, SD_OP_ADVERTISE } , // 1000 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, Half duplex, not advertise.
};
 
// SPEED_1000_FULL_DUPLEX_ARR
static struct speed_duplex_op speed_1000_full_duplex_op_arr[MAX_SPEED_DUPLEX_OPS] = { 
    { SPEED_10_MBPS   , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, Half duplex, not advertise .
    { SPEED_10_MBPS   , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } , // 10 MBPS, full duplex, not advertise.
    { SPEED_100_MBPS  , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 100 MBPS, Half duplex, advertise.
    { SPEED_100_MBPS  , FULL_DUPLEX, SD_OP_NOT_ADVERTISE } ,     // 100 MBPS, full duplex, not advertise.
    { SPEED_1000_MBPS , HALF_DUPLEX, SD_OP_NOT_ADVERTISE } , // 1000 MBPS, Half duplex, not advertise.
    { SPEED_1000_MBPS , FULL_DUPLEX, SD_OP_ADVERTISE } , // 1000 MBPS, full duplex, not advertise.
};
 
 
int mv88e6176_init_speed_duplex_arr(struct speed_duplex_op *ptr ); 
int mv88e6176_init_speed_duplex_arr(struct speed_duplex_op *ptr ) 
{

    if ( ptr == NULL ) 
        return -EINVAL ;

    memcpy((void *)(ptr), &speed_duplex_op_arr[0], sizeof(speed_duplex_op_arr) ) ;

    return 0 ;
}
 
static int mv88e6176_speed_duplex_op( unsigned int port_num, struct speed_duplex_op *ptr ) 
{ 
    int retval = 0 ;
    unsigned int i = 0 ;
    unsigned int reg4_bit_sets = 0 ;
    unsigned int reg9_bit_sets = 0 ;
    unsigned int reg4_prog = 0 ; 
    unsigned int reg9_prog = 0 ; 

    if ( ptr == NULL )
        return -EINVAL ;

    // Read reg 4.
    retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                          MRV_88E6176_PHY_COPPER_AUTO_NEGOTIATION_ADVERTISEMENT_REG , &reg4_bit_sets ) ;
    if ( retval == 0 ) { 
        // Read reg 9.
        retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                          MRV_88E6176_PHY_1000BASE_T_CONTROL_REG , &reg9_bit_sets) ;
        if ( retval == 0 ) { 
            mv88e6176_log(g_debug_print, "--------\n") ;
            for ( i = 0 ; i < MAX_SPEED_DUPLEX_OPS; i++ ) {
                if ( ( ptr->speed == SPEED_10_MBPS) && ( ptr->duplex == HALF_DUPLEX ) && ( ptr->ctrl == SD_OP_ADVERTISE ) ) {
                    // 10, Half duplex, advertise 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_HALF_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_HALF_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_HALF_DUPLEX_VAL_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_10_MBPS) && ( ptr->duplex == HALF_DUPLEX ) && ( ptr->ctrl == SD_OP_NOT_ADVERTISE ) ) { 
                    // 10, Half duplex, no advertise 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_HALF_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_HALF_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_HALF_DUPLEX_VAL_NOT_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_10_MBPS) && ( ptr->duplex == FULL_DUPLEX ) && ( ptr->ctrl == SD_OP_ADVERTISE  ) ) { 
                    // 10, Full duplex, advertise. 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_FULL_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_FULL_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_FULL_DUPLEX_VAL_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_10_MBPS) && ( ptr->duplex == FULL_DUPLEX ) && ( ptr->ctrl == SD_OP_NOT_ADVERTISE ) ) {
                    // 10, Full duplex, Not advertise. 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_FULL_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_FULL_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_10_BASE_TX_FULL_DUPLEX_VAL_NOT_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_100_MBPS) && ( ptr->duplex == HALF_DUPLEX ) && ( ptr->ctrl == SD_OP_ADVERTISE  ) ) { 
                    // 100, Half duplex, advertise 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_HALF_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_HALF_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_HALF_DUPLEX_VAL_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_100_MBPS) && ( ptr->duplex == HALF_DUPLEX ) && ( ptr->ctrl == SD_OP_NOT_ADVERTISE ) ) {
                    // 100, Half duplex, not advertise 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_HALF_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_HALF_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_HALF_DUPLEX_VAL_NOT_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_100_MBPS) && ( ptr->duplex == FULL_DUPLEX ) && ( ptr->ctrl == SD_OP_ADVERTISE  ) ) {
                    // 100, Full duplex, advertise. 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_FULL_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_FULL_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_FULL_DUPLEX_VAL_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_100_MBPS) && ( ptr->duplex == FULL_DUPLEX ) && ( ptr->ctrl == SD_OP_NOT_ADVERTISE ) ) {
                    // 100, Full duplex, advertise. 
                    MV88E6176_SET_FIELD(reg4_bit_sets,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_FULL_DUPLEX_MASK,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_FULL_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_CU_AUTO_NEG_ADVR_REG_100_BASE_TX_FULL_DUPLEX_VAL_NOT_ADVERTISE) ;
                    reg4_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg4_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_1000_MBPS) && ( ptr->duplex == HALF_DUPLEX ) && ( ptr->ctrl == SD_OP_ADVERTISE  ) ) {
                    // 1000, Half duplex, advertise 
                    MV88E6176_SET_FIELD(reg9_bit_sets,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_HALF_DUPLEX_MASK,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_HALF_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_HALF_DUPLEX_VAL_ADVERTISE) ;
                    reg9_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg9_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_1000_MBPS) && ( ptr->duplex == HALF_DUPLEX ) && ( ptr->ctrl == SD_OP_NOT_ADVERTISE ) ) {
                    // 1000, Half duplex, advertise 
                    MV88E6176_SET_FIELD(reg9_bit_sets,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_HALF_DUPLEX_MASK,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_HALF_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_HALF_DUPLEX_VAL_NOT_ADVERTISE) ;
                    reg9_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg9_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_1000_MBPS) && ( ptr->duplex == FULL_DUPLEX ) && ( ptr->ctrl == SD_OP_ADVERTISE  ) ) { 
                    // 1000, Full duplex, advertise 
                    MV88E6176_SET_FIELD(reg9_bit_sets,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_FULL_DUPLEX_MASK,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_FULL_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_FULL_DUPLEX_VAL_ADVERTISE) ;
                    reg9_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg9_bit_sets ) ;
                } else if ( ( ptr->speed == SPEED_1000_MBPS) && ( ptr->duplex == FULL_DUPLEX ) && ( ptr->ctrl == SD_OP_NOT_ADVERTISE ) ) { 
                    // 1000, Full duplex, not advertise 
                    MV88E6176_SET_FIELD(reg9_bit_sets,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_FULL_DUPLEX_MASK,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_FULL_DUPLEX_SHIFT,
                                MRV_88E6176_PHY_1000BASE_T_CONTROL_REG_1000_BASE_T_FULL_DUPLEX_VAL_NOT_ADVERTISE) ;
                    reg9_prog = 1 ; 
                    mv88e6176_log(g_debug_print, "==== speed = %d, duplex = %d, adv = %d , reg_calc = 0x%x \n", ptr->speed, ptr->duplex, ptr->ctrl, reg9_bit_sets ) ;
                } 
                ptr++ ;
            } // End of for loop.
            mv88e6176_log(g_debug_print, "++++++++\n") ;
            // Now write the values back.
            if ( reg4_prog == 1 ) { 
                retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                                 MRV_88E6176_PHY_COPPER_AUTO_NEGOTIATION_ADVERTISEMENT_REG, reg4_bit_sets) ;
                if ( ( retval == 0 ) && ( reg9_prog == 1 ) ) {
                    retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                                 MRV_88E6176_PHY_1000BASE_T_CONTROL_REG, reg9_bit_sets) ;
                }
            }
        }
    }

    return retval ;
}

int mv88e6176_set_port_duplex(int port_num, unsigned int data)
{
    int retval = 0 ;
    unsigned int  reg_val = 0 ;
    unsigned int auto_neg = 0 ;
    unsigned int bits = 0 ;
    unsigned int speed_bits = 0 ;
    unsigned int speed_mbps ;
    struct speed_duplex_op *sp_du_op_ptr=NULL;

    // mutex_lock(&mv88e6176_operation_lock) ;
    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_read_port_status" , &g_operation_lock_name[0])  ;

    if ( ( port_num >= 0 ) && ( port_num <  MRV_88E6176_MAX_PORTS ) ) {
        // Read copper control register.
	retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , &reg_val) ;
        if ( retval == 0 ) {
            // Get autonegotiation value.
            auto_neg = ( reg_val & MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_MASK ) >> MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_SHIFT ;
            if ( auto_neg == MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_VAL_ENABLE_AUTO_NEG_PROCESS ) {
                mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled \n") ;
                // Auto-neg is enabled here.
                // Get present speed.
                // get bits 6 and 13.
                // Get MSB ( bit 6 ).
                bits= (reg_val & MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_MSB_MASK ) >> MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_MSB_SHIFT ; 
                mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, bit 6 = 0x%x \n", bits ) ;
                // Make the bit MSB.
                speed_bits = bits << 1 ;
                // Get LSB ( bit 13) 
                bits = 0 ;
                bits = (reg_val & MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_LSB_MASK ) >> MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_LSB_SHIFT ; 
                mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, bit 13 = 0x%x \n", bits ) ;
                // Now get complete speed : 
                speed_bits = speed_bits | bits ;
                speed_mbps = speed_val[speed_bits] ;
                // We have speed in bits now.
                // Convert to speed in MBPS.
                // If we want to set the duplex value, it will depend on speed. 
                // speed_duplex_arr_slot_t slot = 0 ;
				mv88e6176_log(g_debug_print, "==== In set_port_duplex, port_num = 0x%x,  auto_neg enabled, speed bits = 0x%x , speed_mbps = 0x%x, data = 0x%x \n", port_num, speed_bits, speed_mbps, data ) ;
                if      ( ( speed_mbps == SPEED_10_MBPS ) && ( data == HALF_DUPLEX ) ) {
                    sp_du_op_ptr = &speed_10_half_duplex_op_arr[0] ;
                    mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, SPEED_10, HALF\n") ;
                }
                else if ( ( speed_mbps == SPEED_10_MBPS ) && ( data == FULL_DUPLEX ) ) {
                    sp_du_op_ptr = &speed_10_full_duplex_op_arr[0] ;
                    mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, SPEED_10, FULL\n") ;
                }
                else if ( ( speed_mbps == SPEED_100_MBPS ) && ( data == HALF_DUPLEX ) ) {
                    sp_du_op_ptr = &speed_100_half_duplex_op_arr[0] ;
                    mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, SPEED_100, HALF\n") ;
                }
                else if ( ( speed_mbps == SPEED_100_MBPS ) && ( data == FULL_DUPLEX ) ) {
                    sp_du_op_ptr = &speed_100_full_duplex_op_arr[0] ;
                    mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, SPEED_100, FULL\n") ;
                }
                else if ( ( speed_mbps == SPEED_1000_MBPS ) && ( data == HALF_DUPLEX ) ) {
                    sp_du_op_ptr = &speed_1000_half_duplex_op_arr[0] ;
                    mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, SPEED_1000, HALF\n") ;
                }
                else if ( ( speed_mbps == SPEED_1000_MBPS ) && ( data == FULL_DUPLEX ) ) {
                    sp_du_op_ptr = &speed_1000_full_duplex_op_arr[0] ;
                    mv88e6176_log(g_debug_print, "==== In set_port_duplex, auto_neg enabled, SPEED_1000, FULL\n") ;
                }

                retval = mv88e6176_speed_duplex_op( port_num, sp_du_op_ptr)  ;
                if ( retval == 0 ) { 
                    // Restart phy.
                    reg_val = ( reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_MASK ) | 
                              ( MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_VAL_PHY_RESET << MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_SHIFT ) ;
                    // Now write reg_val into register.
                    retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                                         MRV_88E6176_PHY_COPPER_CTRL_REG, reg_val ) ;
                }
            } else if ( auto_neg == MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_VAL_DISABLE_AUTO_NEG_PROCESS ) { 
                // When Auto negotiation is disabled do the following.
                // Set bit 15 ( Copper reset bit to 1, and Copper Duplex mode bit ( 8 ) 
                // First clean up values in bits 15, and 8 .                        
                reg_val = reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_MASK ;
                reg_val = reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_CU_DUPLEX_MODE_MASK ; 
                // Set bit 15 to PHY reset.
                reg_val = reg_val | ( MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_VAL_PHY_RESET << MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_SHIFT ) ;

                if ( data == HALF_DUPLEX )
                    reg_val = reg_val | ( MRV_88E6176_PHY_CU_CTRL_REG_CU_DUPLEX_MODE_VAL_HALF_DUPLEX << MRV_88E6176_PHY_CU_CTRL_REG_CU_DUPLEX_MODE_SHIFT ) ;
                else if ( data == FULL_DUPLEX )                    
                    reg_val = reg_val | ( MRV_88E6176_PHY_CU_CTRL_REG_CU_DUPLEX_MODE_VAL_FULL_DUPLEX << MRV_88E6176_PHY_CU_CTRL_REG_CU_DUPLEX_MODE_SHIFT ) ;
                // Now write reg_val into register.
                retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                                         MRV_88E6176_PHY_COPPER_CTRL_REG, reg_val ) ;
            }
        } else {
            retval = -EINVAL ;
        }
    } else  {
        retval = -EINVAL ;
    }

    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_read_port_status" , &g_operation_lock_name[0])  ;
    return retval ;
}


int mv88e6176_set_port_speed(int port_num, unsigned int speed)
{
    int retval = 0 ;
    unsigned int  reg_val = 0 ;
    unsigned int auto_neg = 0 ;
    struct speed_duplex_op *sp_du_op_ptr = NULL ;
    unsigned int dup = 0 ;
    unsigned int rval = 0 ;
    duplex_t duplex;

    // mutex_lock(&mv88e6176_operation_lock) ;
    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_read_port_status" , &g_operation_lock_name[0])  ;

    if ( ( port_num >= 0 ) && ( port_num <  MRV_88E6176_MAX_PORTS ) ) {
        // Read copper control register.
	retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , &reg_val) ;
        if ( retval == 0 ) {
            // Get autonegotiation value.
            auto_neg = ( reg_val & MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_MASK ) >> MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_SHIFT ;
            if ( auto_neg == MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_VAL_ENABLE_AUTO_NEG_PROCESS ) {
                // Auto-neg is enabled here.
                // Speed is not set in any registers as such. We don't advertise values other than the speed 
                // value we want to set and that will take care of speed setting. This is done because autoneg is enabled.
                // Get duplex information.
                retval = mv88e6176_port_glob_read_reg((port_num + MRV_88E6176_PORT_START_DEV_ADDR), 
                                                      MRV_88E6176_PORT_STATUS_REG_ADDR, &rval)  ;
                if ( retval == 0 ) {
                    dup = (rval & MRV_88E6176_PORT_STATUS_REG_DUPLEX_MASK) >> MRV_88E6176_PORT_STATUS_REG_DUPLEX_SHIFT ;
                    // If dup is 0 - Half duplex.
                    // If dup is 1 - Full duplex.
                    duplex = (duplex_t)dup ;
                    if ( duplex == FULL_DUPLEX ) { 
                        switch ( speed ) {
                            case SPEED_10_MBPS : 
                            {
                                sp_du_op_ptr = &speed_10_full_duplex_op_arr[0] ;
                                break ;
                            }
                            case SPEED_100_MBPS : 
                            {
                                sp_du_op_ptr = &speed_100_full_duplex_op_arr[0] ;
                                break ;
                            }
                            case SPEED_1000_MBPS : 
                            {
                                sp_du_op_ptr = &speed_1000_full_duplex_op_arr[0] ;
                                break ;
                            }
                            default : 
                            {
                                retval = -EINVAL ;
                                break ;
                            }
                        }
                    } else if ( duplex == HALF_DUPLEX ) { 
                        switch ( speed ) {
                            case SPEED_10_MBPS : 
                            {
                                sp_du_op_ptr = &speed_10_half_duplex_op_arr[0] ;
                                break ;
                            }
                            case SPEED_100_MBPS : 
                            {
                                sp_du_op_ptr = &speed_100_half_duplex_op_arr[0] ;
                                break ;
                            }
                            case SPEED_1000_MBPS : 
                            {
                                sp_du_op_ptr = &speed_1000_half_duplex_op_arr[0] ;
                                break ;
                            }
                            default : 
                            {
                                retval = -EINVAL ;
                                break ;
                            }
                        }
                    }
                        
                    // Set new speed with correct duplex and reset the phy only if all settings have succeeded.
                    if ( sp_du_op_ptr != NULL ) {
                        retval = mv88e6176_speed_duplex_op( port_num, sp_du_op_ptr)  ;
                    }

                    if ( retval == 0 ) { 
                        // 3. Restart auto negotiation process.
            	        retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                                                             MRV_88E6176_PHY_COPPER_CTRL_REG , &reg_val) ;
                        if ( retval == 0 ) {
                            // 1. Restart auto-neg setting.
                            reg_val = ( reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_RESTART_CU_AUTO_NEGOTIATION_MASK ) | 
                                      ( MRV_88E6176_PHY_CU_CTRL_REG_RESTART_CU_AUTO_NEG_VAL_RESTART_AUTO_NEG_PROCESS << MRV_88E6176_PHY_CU_CTRL_REG_RESTART_CU_AUTO_NEGOTIATION_SHIFT ) ;
                            // 2. Apart from restarting auto negotiation process, set correct speed val for bits 13,6.
                            //    When auto neg is on setting advertized values is enough, but setting correct speed 
                            //    value for bits 13,6 is useful because if auto-neg is turned off, we are not in a weird speed state.
                            reg_val = reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_LSB_MASK ; 
                            reg_val = reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_MSB_MASK ; 
                            switch ( speed ) {
                            case SPEED_10_MBPS : 
                            {
                                // Nothing to set, they are already 0, cleared up outside of switch statement.
                                break ;
                            }
                            case SPEED_100_MBPS : 
                            {
                                // Set bit 13 to 1.
                                reg_val = reg_val | 0x2000 ;
                                break ;
                            }
                            case SPEED_1000_MBPS : 
                            {
                                // Set bit 6 to 1.
                                reg_val = reg_val | 0x40 ;
                                break ;
                            }
                            default : 
                                break ;
                            } // End of switch. 

                            retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                                                          MRV_88E6176_PHY_COPPER_CTRL_REG, reg_val) ;
                        } else {
                            // Restarting auto neg process failed.
                            retval = -EINVAL ;
                        }
                    } else {
                        // Setting new DUPLEX value failed.
                        retval = -EINVAL ;
                    }
                } else {
                    // Reading DUPLEX status failed.
                    retval = -EINVAL ;
                }
            } else if ( auto_neg == MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_VAL_DISABLE_AUTO_NEG_PROCESS ) { 
                // When Auto negotiation is disabled do the following.
                // Set bit 15 ( Copper reset bit to 1, and speed bits according to required values ( bits 6, 13 ) 
                // First clean up values in bits 15, 13, and 6 .                        
                reg_val = reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_MASK ;
                reg_val = reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_LSB_MASK ; 
                reg_val = reg_val & ~MRV_88E6176_PHY_CU_CTRL_REG_SPEED_SELECTION_MSB_MASK ; 
                // Set bit 15 to PHY reset.
                reg_val = reg_val | MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_VAL_PHY_RESET ;

                //  +-------+-------+--------+
                //  | Speed | Bit 6 | BIT 13 | 
                //  |       | (MSB) | (LSB)  | 
                //  +-------+-------+--------+
                //  |  10   |    0  |    0   |
                //  +-------+-------+--------+
                //  |  100  |    0  |    1   |
                //  +-------+-------+--------+
                //  |  1000 |    1  |    0   |
                //  +-------+-------+--------+
                //  |UNKNOWN|    1  |    1   |
                //  +-------+-------+--------+

                switch ( speed ) {
                case SPEED_10_MBPS : 
                {
                    // Nothing to set, they are already 0, cleared up outside of switch statement.
                    break ;
                }
                case SPEED_100_MBPS : 
                {
                    // Set bit 13 to 1.
                    reg_val = reg_val | 0x2000 ;
                    break ;
                }
                case SPEED_1000_MBPS : 
                {
                    // Set bit 6 to 1.
                    reg_val = reg_val | 0x40 ;
                    break ;
                }
                default : 
                    break ;
                } // End of switch. 
                // Now write reg_val into register.
                retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , 
                                         MRV_88E6176_PHY_COPPER_CTRL_REG, reg_val ) ;
            }
        } else {
            retval = -EINVAL ;
        }
    } else  {
        retval = -EINVAL ;
    }

    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_read_port_status" , &g_operation_lock_name[0])  ;
    return retval ;
}

int mv88e6176_read_port_physical_control(int port_num, PORT_DETS_T pd , unsigned int *data)
{
    int retval = 0 ;
    unsigned int rval = 0 ;

    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_read_port_physical_control" , &g_operation_lock_name[0])  ;

    if ( ( port_num >= 0 ) && ( port_num <  MRV_88E6176_MAX_PORTS ) && ( data != NULL ) ) {
        retval = mv88e6176_port_glob_read_reg((port_num + MRV_88E6176_PORT_START_DEV_ADDR), MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_ADDR, &rval)  ;
        if ( retval == 0 ) {
            switch (pd) {
                case FCVALUE :
                {
                    *data = (rval & MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FCVALUE_MASK ) >> MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FCVALUE_SHIFT ;
                    break ;
                }
                case FORCEDFC :
                {
                    *data = (rval & MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FORCEDFC_MASK ) >> MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FORCEDFC_SHIFT ;
                    break ;
                }
                default :
                {
                    printk("%s Invalid type %d \n", __FUNCTION__, pd ) ;
                    break ;
                }
            } // End of switch
        } else {
            retval = -EINVAL ;
        }
    } else  {
        retval = -EINVAL ;
    }

    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_read_port_physical_status" , &g_operation_lock_name[0])  ;
    return retval ;
}

int mv88e6176_write_port_physical_control(int port_num, PORT_DETS_T pd , unsigned int data)
{
    int retval = 0 ;
    unsigned int rval = 0 ;
    unsigned int wval = 0 ;

    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_read_port_physical_control" , &g_operation_lock_name[0])  ;

    if ( ( port_num >= 0 ) && ( port_num <  MRV_88E6176_MAX_PORTS ) ) {
        // retval = mv88e6176_read_reg((port_num + MRV_88E6176_PORT_START_DEV_ADDR), MRV_88E6176_PORT_STATUS_REG_ADDR, &rval)  ;
        retval = mv88e6176_port_glob_read_reg((port_num + MRV_88E6176_PORT_START_DEV_ADDR), MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_ADDR, &rval)  ;
        if ( retval == 0 ) {
            switch (pd) {
                case FCVALUE :
                {
                    wval = (rval & ( ~MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FCVALUE_MASK ) ) | ( data << MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FCVALUE_SHIFT ) ;
                    break ;
                }
                case FORCEDFC :
                {
                    wval = (rval & ( ~MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FORCEDFC_MASK ) ) | ( data << MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_FORCEDFC_SHIFT ) ;
                    break ;
                }
                default :
                {
                    printk("%s Invalid type %d \n", __FUNCTION__, pd ) ;
                    break ;
                }
            } // End of switch
            // Now write the value to register.
            retval = mv88e6176_port_glob_write_reg((port_num + MRV_88E6176_PORT_START_DEV_ADDR), MRV_88E6176_PORT_PHYSICAL_CONTROL_REG_ADDR, wval)  ;
            if ( retval == 0 ) {
                printk(" Wrote 0x%x into port physical control reg. rval was 0x%x \n", wval, rval ) ;
            }
        } else {
            retval = -EINVAL ;
        }
    } else  {
        retval = -EINVAL ;
    }

    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_read_port_physical_status" , &g_operation_lock_name[0])  ;
    return retval ;
}

int mv88e6176_autonegotiation_details(unsigned int port_num, unsigned int *enabled, unsigned int *completed , unsigned int *phy_addr , unsigned int op, unsigned int wval)
{
    int retval = 0 ;
    unsigned int temp = 0 ;

    if ( ( port_num >=  MRV_88E6176_MAX_PORTS ) || ( enabled == NULL ) || ( completed == NULL ) || ( phy_addr == NULL ) )
        return -EINVAL ;

    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_autonegotiation_details" , &g_operation_lock_name[0])  ;

    // Port numbers and Phys are mapped one to one.
    // Port 0 is associated to Phy 0, etc.
    // But there are only 5 phys and 7 ports.
    // ports 5, 6 don't have a phy associated to them.
    if ( port_to_phy_dev_addr_map[port_num] == MRV_88E6176_PHY_UNKNOWN_ADDR ) {
	*phy_addr = MRV_88E6176_PHY_UNKNOWN_ADDR ;
	retval = -EINVAL ;
    } else {
        *phy_addr = port_to_phy_dev_addr_map[port_num] ;

        if ( op == AUTO_NEG_SHOW ) {
	    retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , &temp) ;
	    // printk("=== ENABLED val = 0x%x \n", temp ) ;
	    if ( retval == 0 ) {
		*enabled = (temp & MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_MASK) >>  MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_SHIFT ;
		temp = 0 ;
		retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_STATUS_REG , &temp) ;
		// printk("=== COMPLETED val = 0x%x \n", temp ) ;
		if ( retval == 0 ) {
		    *completed = (temp & MRV_88E6176_PHY_CU_STATUS_REG_CU_AUTO_NEG_COMPLETE_MASK) >> MRV_88E6176_PHY_CU_STATUS_REG_CU_AUTO_NEG_COMPLETE_SHIFT ;
		}
	    }
        }
        else if ( op == AUTO_NEG_STORE ) {
	    retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , &temp) ;
	    if ( retval == 0 ) {
                // Set Auto neg bit.
		temp = ( temp & ~MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_MASK ) | ( wval <<  MRV_88E6176_PHY_CU_CTRL_REG_AUTO_NEGOTIATION_ENABLE_SHIFT ) ;
                // Restart auto negotiation.
                temp = temp | MRV_88E6176_PHY_CU_CTRL_REG_COPPER_RESET_MASK ;
                temp = ( temp & ~MRV_88E6176_PHY_CU_CTRL_REG_RESTART_CU_AUTO_NEGOTIATION_MASK ) | 
                       ( MRV_88E6176_PHY_CU_CTRL_REG_RESTART_CU_AUTO_NEG_VAL_RESTART_AUTO_NEG_PROCESS << MRV_88E6176_PHY_CU_CTRL_REG_RESTART_CU_AUTO_NEGOTIATION_SHIFT ) ;
                // Write the new value.                 
                retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , temp) ;
            }
        }
    }
    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_autonegotiation_details" , &g_operation_lock_name[0])  ;
    return retval ;
}

int mv88e6176_power_down(unsigned int port_num, unsigned int *pd, unsigned int *phy_addr , unsigned int op, unsigned int wval)
{
    int retval = 0 ;
    unsigned int temp = 0 ;

    if ( ( port_num >=  MRV_88E6176_MAX_PORTS ) || ( pd == NULL ) || ( phy_addr == NULL ) )
        return -EINVAL ;

    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_autonegotiation_details" , &g_operation_lock_name[0])  ;

    // Port numbers and Phys are mapped one to one.
    // Port 0 is associated to Phy 0, etc.
    // But there are only 5 phys and 7 ports.
    // ports 5, 6 don't have a phy associated to them.
    if ( port_to_phy_dev_addr_map[port_num] == MRV_88E6176_PHY_UNKNOWN_ADDR ) {
	*phy_addr = MRV_88E6176_PHY_UNKNOWN_ADDR ;
	retval = -EINVAL ;
    } else {
        *phy_addr = port_to_phy_dev_addr_map[port_num] ;

        if ( op == POWER_DOWN_SHOW ) {
	    retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , &temp) ;
	    // printk("=== ENABLED val = 0x%x \n", temp ) ;
	    if ( retval == 0 ) {
		*pd = temp & MRV_88E6176_PHY_CU_CTRL_REG_POWER_DOWN_MASK >>  MRV_88E6176_PHY_CU_CTRL_REG_POWER_DOWN_SHIFT ;
	    }
        }
        else if ( op == POWER_DOWN_STORE ) {
	    retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , &temp) ;
	    if ( retval == 0 ) {
		temp = ( temp & ~MRV_88E6176_PHY_CU_CTRL_REG_POWER_DOWN_MASK ) | ( wval <<  MRV_88E6176_PHY_CU_CTRL_REG_POWER_DOWN_SHIFT ) ;
                retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_PHY_COPPER_CTRL_REG , temp) ;

                // Now write into register 16, page 0 also.
	        retval = mv88e6176_phy_read_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 , MRV_88E6176_COPPER_SPECIFIC_CONTROL_REG1_REG_ADDR , &temp) ;
	        if ( retval == 0 ) {
                    temp = ( temp & ~MRV_88E6176_PHY_CU_SPECIFIC_CONTROL_REG1_POWER_DOWN_MASK ) |
                           ( wval <<  MRV_88E6176_PHY_CU_SPECIFIC_CONTROL_REG1_POWER_DOWN_SHIFT ) ;
                    retval = mv88e6176_phy_write_reg( port_to_phy_dev_addr_map[port_num], MRV_88E6176_PAGE_0 ,
                                                      MRV_88E6176_COPPER_SPECIFIC_CONTROL_REG1_REG_ADDR , temp) ;
                }
            }
        }
    }
    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_autonegotiation_details" , &g_operation_lock_name[0])  ;
    return retval ;
}

static int mv88e6176_stat_op(unsigned int wval)
{
    int retval = 0 ;
    unsigned int rval = 0 ;
    unsigned char iter = 0 ;
    unsigned char busybit = 1 ;

    retval = mv88e6176_port_glob_write_reg(MRV_88E6176_GLOBAL1_DEV_ADDR, MRV_88E6176_GLOBAL1_STATS_OP_REG_ADDR, wval) ;
    if ( retval == 0 ) {
        // Wait until the Stats Busy bit becomes 0. Don't wait indefinitely.
        while ( ( iter < STATS_BUSY_READ_RETRY ) && ( busybit == 1 ) && ( retval == 0 ) ) {
            rval = 0 ;
            retval = mv88e6176_port_glob_read_reg( MRV_88E6176_GLOBAL1_DEV_ADDR, MRV_88E6176_GLOBAL1_STATS_OP_REG_ADDR,&rval) ;
            busybit = ( rval & MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_BUSY_MASK ) >> MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_BUSY_SHIFT ;
            udelay(STATS_BUSY_READ_WAIT) ;
            iter++ ;
        } // End of while loop.

        if ( ( busybit ) || ( retval != 0 ) ) {
            printk("%s:: busybit is still set, or retval != 0, rval = %d, retval = 0x%x \n", __FUNCTION__, rval, retval ) ;
            retval = -EINVAL ; // Return error here.
        }
    }
    return retval ;
}

// Read statistics command. To be called after a capture command.
static int mv88e6176_read_cmd(unsigned int port_num, unsigned int stat)
{
    unsigned short wval = 0 ;
    // set Statsop = 4, StatsPort = port #, Stats Ptr = Counter name.
    wval = ( 1 << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_BUSY_SHIFT ) |
           ( MRV_88E6876_GLOBAL1_STATS_OP_REG_HISTOGRAM_MODE_VAL_CNT_RX_TX_FRAMES << MRV_88E6876_GLOBAL1_STATS_OP_REG_HISTOGRAM_MODE_SHIFT ) |
           ( MRV_88E6176_GLOBAL1_STATS_OP_REG_STATS_OP_VAL_READ_A_CAPTURED_OR_DIRECT_COUNTER << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_OP_SHIFT ) |
           ( (port_num+1) << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_PORT_SHIFT ) |
           ( stat << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_PTR_SHIFT ) ;

    return mv88e6176_stat_op(wval) ;
}

// Capture statistics.
static int mv88e6176_capture_cmd(unsigned int port_num, unsigned int stat)
{
    unsigned short wval = 0 ;
    // set Statsop = 5, StatsPort = port #, Stats Ptr = Counter name.
    wval = ( 1 << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_BUSY_SHIFT ) |
           ( MRV_88E6876_GLOBAL1_STATS_OP_REG_HISTOGRAM_MODE_VAL_CNT_RX_TX_FRAMES << MRV_88E6876_GLOBAL1_STATS_OP_REG_HISTOGRAM_MODE_SHIFT ) |
           ( MRV_88E6176_GLOBAL1_STATS_OP_REG_STATS_OP_VAL_CAPTURES_ALL_COUNTERS_FOR_A_PORT << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_OP_SHIFT ) |
           ( (port_num+1) << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_PORT_SHIFT ) |
           ( stat << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_PTR_SHIFT ) ;

    return mv88e6176_stat_op(wval) ;
}

int mv88e6176_clear_port_stats(unsigned int port_num)
{
    unsigned short wval = 0 ;
    int retval = 0 ;
    // set Statsop = 5, StatsPort = port #, Stats Ptr = Counter name.
    wval = ( 1 << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_BUSY_SHIFT ) |
           ( MRV_88E6176_GLOBAL1_STATS_OP_REG_STATS_OP_VAL_FLUSH_ALL_COUNTERS_FOR_A_PORT << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_OP_SHIFT ) |
           ( (port_num+1) << MRV_88E6876_GLOBAL1_STATS_OP_REG_STATS_PORT_SHIFT ) ;

    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_clear_port_stats", &g_operation_lock_name[0])  ;
    retval = mv88e6176_stat_op(wval) ;
    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_clear_port_stats", &g_operation_lock_name[0])  ;
    return retval ;
}

int mv88e6176_read_port_stats(unsigned int port_num, unsigned int stat, unsigned int *val )
{
    int retval = 0 ;
    unsigned int rval = 0 ;

    if ( val == NULL )
        return -EINVAL ;

    mv88e6176_lock(mv88e6176_operation_lock, flags, g_operation_lock_cnt, "mv88e6176_read_port_stats", &g_operation_lock_name[0])  ;

    //1. Send Capture command.
    retval = mv88e6176_capture_cmd(port_num, stat) ;
    if ( retval == 0 ) {

        //2. Send Read command.
        retval = mv88e6176_read_cmd(port_num, stat) ;
        if ( retval == 0 ) {

            //3. Read the data.
            *val = 0 ;
            retval = mv88e6176_port_glob_read_reg( MRV_88E6176_GLOBAL1_DEV_ADDR, MRV_88E6176_GLOBAL1_STATS_COUNTER_BYTES_3_2_REG_ADDR,&rval) ;
            if ( retval == 0 ) {
                *val = rval << 16 ;
                rval = 0 ;
                retval = mv88e6176_port_glob_read_reg( MRV_88E6176_GLOBAL1_DEV_ADDR, MRV_88E6176_GLOBAL1_STATS_COUNTER_BYTES_1_0_REG_ADDR,&rval) ;
                if ( retval == 0 ) {
                    *val = *val | rval ;
                }
            } // Read Bytes 3,2
        } // Read Cmd
    } // Capture cmd.

    mv88e6176_unlock(mv88e6176_operation_lock, flags, g_operation_unlock_cnt, "mv88e6176_read_port_stats", &g_operation_lock_name[0])  ;

    return retval ;
}

/* Device structure */
static struct file_operations mv88e6176_fops = {
    unlocked_ioctl:		mv88e6176_ioctl,
    open:		mv88e6176_open,
};


/* Module Init Function*/
static int mv88e6176_dev_init(void)
{
    int err = 0 ;
    //  Register char dev.
    if ((err = register_chrdev(MV88E6176_MAJOR, MV88E6176_DEV_NAME, &mv88e6176_fops)) < 0)
    {
        printk ("MV88e6176: failed to register device (err %d)\n", err);
        return -ENODEV;
    } else {
        printk ("MV88e6176: Success int registering mv88e6176 device \n");
    }
    return 0;
}

/* Module Exit Function*/
void mv88e6176_dev_exit(void)
{
    /* unregister the device */
    unregister_chrdev(MV88E6176_MAJOR, MV88E6176_DEV_NAME);
    printk(KERN_ALERT "MV88E6176 module unloaded sucessfully.\n") ;
}

/* Open */
static int mv88e6176_open(struct inode * inode, struct file * filp)
{
    filp->f_op =&mv88e6176_fops;
    return  0;
}

/* ioctl */
static long  mv88e6176_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    enet_switch_ioctl_t   pkt;
    int retval = 0 ;

    if ( copy_from_user((void*)&pkt,(void*)arg,sizeof(enet_switch_ioctl_t))) {
        return -EFAULT;
    }

    switch ( cmd )
    {
        case ENET_SWITCH_READ_REG:
            if ( ( pkt.dev_addr >= MRV_88E6176_PHY_START_DEV_ADDR ) && ( pkt.dev_addr <= MRV_88E6176_PHY_END_DEV_ADDR ) ) {
                retval = mv88e6176_phy_read_reg(pkt.dev_addr, pkt.page_addr, pkt.reg_addr, &(pkt.data) ) ;
                // printk("=== ENET_SWITCH_READ_REG ioctl, PHY data = 0x%x \n", pkt.data ) ;
            } else if ( ( ( pkt.dev_addr >= MRV_88E6176_PORT_START_DEV_ADDR ) && ( pkt.dev_addr <= MRV_88E6176_PORT_END_DEV_ADDR ) ) ||
                        ( pkt.dev_addr == MRV_88E6176_GLOBAL1_DEV_ADDR ) ||
                        ( pkt.dev_addr == MRV_88E6176_GLOBAL2_DEV_ADDR ) ) {
                retval = mv88e6176_port_glob_read_reg(pkt.dev_addr, pkt.reg_addr, &(pkt.data) ) ;
                // printk("=== ENET_SWITCH_READ_REG ioctl, PORT  data = 0x%x \n", pkt.data ) ;
            }

            if ( retval == 0 ) {
                if (copy_to_user((void*)arg,(void*) &pkt , sizeof(enet_switch_ioctl_t)))
                    return -EFAULT ;
            }
        break;

        case ENET_SWITCH_WRITE_REG:

            if ( ( pkt.dev_addr >= MRV_88E6176_PHY_START_DEV_ADDR ) && ( pkt.dev_addr <= MRV_88E6176_PHY_END_DEV_ADDR ) ) {
                retval = mv88e6176_phy_write_reg(pkt.dev_addr, pkt.page_addr, pkt.reg_addr, pkt.data ) ;
                // printk("=== ENET_SWITCH_READ_REG ioctl, PHY data = 0x%x \n", pkt.data ) ;
            } else if ( ( ( pkt.dev_addr >= MRV_88E6176_PORT_START_DEV_ADDR ) && ( pkt.dev_addr <= MRV_88E6176_PORT_END_DEV_ADDR ) ) ||
                        ( pkt.dev_addr == MRV_88E6176_GLOBAL1_DEV_ADDR ) ||
                        ( pkt.dev_addr == MRV_88E6176_GLOBAL2_DEV_ADDR ) ) {
                retval = mv88e6176_port_glob_write_reg(pkt.dev_addr, pkt.reg_addr, pkt.data ) ;
                // printk("=== ENET_SWITCH_PORT_GLOB_WRITE_REG ioctl, PORT  data = 0x%x \n", pkt.data ) ;
            }
        break;

        default:
            break;
    }
    return retval ;
}

int __init
mv88e6176_init(void)
{
    int retval = 0 ;

    retval = mv88e6176_dev_init() ;
    if ( retval == 0 )
        retval = sw_sysfs_init() ;
    return retval ;
}

static void
mv88e6176_exit(void)
{
    printk ( "mv88e6176 Driver is unloaded sucessfully.\n");
    return;
}

module_init(mv88e6176_init);
module_exit(mv88e6176_exit);

MODULE_DESCRIPTION("Marvell 88E6176 Driver");
MODULE_AUTHOR("NetApp Inc");
MODULE_LICENSE("GPL");

